﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Json;
using System.Text;
using System.Text.RegularExpressions;

namespace rbcom {

    public class RubyProgramManager {

        private static string RUBY18_VERSION_PATTERN = @"^(?<name>\w+)\s+(?<version>\d+\.\d+\.\d+)\s+\((?<verdate>\S+)\s+(\w+)\s+(?<patchlevel>\d+)\)\s+\[((?<arch>\w+)\-)?(?<platform>\w+)\].*$";
        private static string RUBY19_VERSION_PATTERN = @"^(?<name>\w+)\s+(?<version>\d+\.\d+\.\d+)[pP](?<patchlevel>\d+)\s+\((?<verdate>\S+)\)\s+\[((?<arch>\w+)\-)?(?<platform>\w+)\].*$";
        private static string JRUBY_VERSION_PATTERN = @"^(?<name>\w+)\s+(?<version>\d+\.\d+\.\d+)\s+\((?<compatver>\S+)\)\s+(?<verdate>\S+)\s+(?<platform>.*?)\s+\[(?<os>.*?)\-(?<arch>\w+)\].*$";

        public List<RubyProgram> Rubies { get; private set; }

        public RubyProgramManager() {
            this.Rubies = new List<RubyProgram>();
        }

        /// <summary>
        /// Ruby名を解析します. Ruby名は
        /// 
        /// [ruby-]x.y.z[-pddd][-platform]
        /// 
        /// 形式です
        /// </summary>
        /// <param name="name">Ruby名</param>
        /// <param name="arch">アーキテクチャ</param>
        /// <returns></returns>
        public static RubyProgram ParseRubyVersion(string name, string arch = null) {
            string pattern = @"((?<name>\w+)\-)?(?<version>\d+(\.\d+){0,2}(\-?[pP](?<patch>\d+))?)(\-(?<platform>\S+))?";
            Match match = Regex.Match(name, pattern);

            if (match.Success) {
                RubyProgram ruby = new RubyProgram();
                if (match.Groups[@"name"].Success) {
                    ruby.Name = match.Groups[@"name"].Value;
                }
                if (match.Groups[@"version"].Success) {
                    ruby.Version = new RubyVersion(match.Groups[@"version"].Value);
                }
                if (match.Groups[@"platform"].Success) {
                    ruby.Platform = match.Groups[@"platform"].Value;
                }
                if (arch != null) {
                    ruby.Architecture = arch;
                }

                return ruby;
            } else {
                return null;
            }
        }

        /// <summary>
        /// 左側Rubyに右側Rubyがマッチするかを調べます
        /// </summary>
        /// <param name="left">比較条件となるRubyProgram</param>
        /// <param name="right">比較されるRubyProgram</param>
        /// <returns>マッチしているならtrue</returns>
        public static bool MatchRubyProgram(RubyProgram left, RubyProgram right) {
            if (left == null || right == null) {
                return false;
            }

            bool matched = true;
            bool tested = false;

            if (left.Architecture != null && left.Architecture.Length > 0) {
                if (right.Architecture != left.Architecture) {
                    Debug.WriteLine("[DEBUG] Architecture Unmatched: {0} <=> {1}", left.Architecture, right.Architecture);
                    matched = false;
                } else {
                    Debug.WriteLine("[DEBUG] Architecture Matched: {0} <=> {1}", left.Architecture, right.Architecture);
                }
                tested = true;
            } else {
                Debug.WriteLine("[NOTICE] Skip Architecture Matching");
            }
            if (matched && left.Platform != null && left.Platform.Length > 0) {
                if (right.Platform != left.Platform) {
                    Debug.WriteLine("[DEBUG] Platform Unmatched: {0} <=> {1}", left.Platform, right.Platform);
                    matched = false;
                } else {
                    Debug.WriteLine("[DEBUG] Platform Matched: {0} <=> {1}", left.Platform, right.Platform);
                }
                tested = true;
            } else {
                Debug.WriteLine("[NOTICE] Skip Platform Matching");
            }
            if (matched && left.Version != null) {
                if (!left.Version.Match(right.Version)) {
                    Debug.WriteLine("[DEBUG] Version Unmatched: {0} <=> {1}", left.Version, right.Version);
                    matched = false;
                } else {
                    Debug.WriteLine("[DEBUG] Version Matched: {0} <=> {1}", left.Version, right.Version);
                }
                tested = true;
            } else {
                Debug.WriteLine("[NOTICE] Skip Version Matching");
            }
            if (matched && left.Name != null && left.Name.Length > 0) {
                Match nmMatch = Regex.Match(right.Name, left.Name, RegexOptions.IgnoreCase);
                if (nmMatch.Success) {
                    Debug.WriteLine("[DEBUG] Name Matched: {0} <=> {1}", left.Name, right.Name);
                } else {
                    Debug.WriteLine("[DEBUG] Name Unmatched: {0} <=> {1}", left.Name, right.Name);
                    matched = false;
                }
                tested = true;
            } else {
                Debug.WriteLine("[NOTICE] Skip Name Matching");
            }

            if (!tested && matched) {
                matched = false;
            }

            return matched;
        }

        /// <summary>
        /// 新しくRubyを管理下に入れます.
        /// </summary>
        /// <param name="path">ruby.exeへのパス</param>
        /// <returns>成功時true, rubyではない場合や、バージョンを解析出来なかった場合はfalse</returns>
        public RubyProgram Add(string path) {
            if (!File.Exists(path)) {
                throw new FileNotFoundException("指定ファイルは存在しません", path);
            }
            FileAttributes attributes = File.GetAttributes(path);
            if ((attributes & FileAttributes.Directory) == FileAttributes.Directory) {
                throw new FileNotFoundException("ディレクトリが指定されています", path);
            }

            // バージョンを取得
            string fileName = Path.GetFileName(path);
            string fileLocation = Path.GetDirectoryName(path);
            Process proc = new Process();
            proc.StartInfo.CreateNoWindow = false;
            proc.StartInfo.RedirectStandardError = false;
            proc.StartInfo.RedirectStandardInput = false;
            proc.StartInfo.RedirectStandardOutput = true;
            proc.StartInfo.UseShellExecute = false;
            proc.StartInfo.FileName = path;
            proc.StartInfo.Arguments = @"--version";
            proc.Start();
            string result = proc.StandardOutput.ReadToEnd();
            proc.WaitForExit();

            // バージョンを解析
            string[] patterns = new string[]{RUBY19_VERSION_PATTERN, RUBY18_VERSION_PATTERN, JRUBY_VERSION_PATTERN};
            RubyProgram ruby = null;
            foreach (string pattern in patterns) {
                Match match = Regex.Match(result, pattern);
                if (match.Success) {
                    Debug.WriteLine("[DEBUG] Command  = {0}", match.Groups[@"name"]);
                    Debug.WriteLine("[DEBUG] Version  = {0}", match.Groups[@"version"]);
                    Debug.WriteLine("[DEBUG] PatchLv  = {0}", match.Groups[@"patchlevel"]);
                    Debug.WriteLine("[DEBUG] Arch     = {0}", match.Groups[@"arch"]);
                    Debug.WriteLine("[DEBUG] Platform = {0}", match.Groups[@"platform"]);

                    string rbName = match.Groups[@"name"].Value;
                    string rbVer = match.Groups[@"version"].Value;
                    string rbArch = match.Groups[@"arch"].Value;
                    string rbPlatform = match.Groups[@"platform"].Value;

                    ruby = new RubyProgram();
                    ruby.Command = fileName;
                    ruby.Location = fileLocation;
                    ruby.Name = rbName;
                    ruby.Version = new RubyVersion(rbVer);
                    if (match.Groups[@"patchlevel"].Success) {
                        int rbPatchValue = 0;
                        string rbPatch = match.Groups[@"patchlevel"].Value;
                        Int32.TryParse(rbPatch, out rbPatchValue);
                        ruby.Version.PatchLevel = rbPatchValue;
                    }
                    ruby.Architecture = rbArch;
                    ruby.Platform = rbPlatform;

                    // JRuby判定
                    if (rbPlatform != null) {
                        foreach (string part in rbPlatform.Split(new char[]{' '})) {
                            if (part.ToLower() == @"java") {
                                ruby.Platform = @"java";
                                break;
                            }
                        }
                    }

                    break;
                }
            }

            // 追加
            if (ruby != null) {
                this.Rubies.Add(ruby);
            }

            return ruby;
        }

        /// <summary>
        /// 指定したRubyに一致したRubyを管理下から除外します
        /// </summary>
        /// <param name="target">比較するRuby</param>
        /// <returns>削除されたRubyプログラムリスト</returns>
        public List<RubyProgram> Delete(RubyProgram target) {
            List<RubyProgram> removed = new List<RubyProgram>();
            List<RubyProgram> keep = new List<RubyProgram>();

            foreach (RubyProgram ruby in this.Rubies) {
                if (MatchRubyProgram(ruby, target)) {
                    // マッチしたので削除
                    removed.Add(ruby);
                } else {
                    // マッチしていないので残す
                    keep.Add(ruby);
                }
            }

            // 残すもので置き換え
            this.Rubies = keep;

            // 消したものを返す
            return removed;
        }

        /// <summary>
        /// 指定したバージョン/エイリアスにマッチしたRubyを管理下から除外します.
        /// エイリアスを検索しないため、条件は#Findよりも厳しくなります
        /// </summary>
        /// <param name="name">Rubyバージョン</param>
        /// <returns>削除されたRubyプログラムリスト</returns>
        public List<RubyProgram> Delete(string name) {
            return this.Delete(ParseRubyVersion(name));
        }

        /// <summary>
        /// 指定したバージョン及びエイリアスマッチしたRubyを探します
        /// </summary>
        /// <param name="name">探す名称. 形式は#Parseを参照</param>
        /// <param name="aliases">エイリアスを検索対象に含めるかどうか (Default:false)</param>
        /// <returns>見つかったRubyのリスト</returns>
        public List<RubyProgram> Find(string name, bool aliases = false) {
            List<RubyProgram> rubies = new List<RubyProgram>();

            // 検索用オブジェクト
            RubyProgram target = ParseRubyVersion(name);

            Debug.WriteLineIf(target != null, String.Format("[DEBUG] Target Version: {0}", target));
            Debug.WriteLineIf(target == null, String.Format("[DEBUG] Target Version: {0}", name));

            foreach (RubyProgram ruby in this.Rubies) {
                Debug.WriteLine("[DEBUG] Matching to {0}", ruby);

                // オブジェクト同士で比較
                bool matched = MatchRubyProgram(target, ruby);

                // エイリアスチェック
                if (!matched && aliases && ruby.Aliases != null && name.Length > 0) {
                    Debug.WriteLine("[DEBUG] Checking Aliases");

                    foreach (String alias in ruby.Aliases.ToArray()) {
                        Match aliasMatch = Regex.Match(alias, name, RegexOptions.IgnoreCase);
                        if (aliasMatch.Success) {
                            matched = true;
                            break;
                        }
                    }
                } else {
                    Debug.WriteLine("[NOTICE] Skip Alias Matching");
                }

                // 名前の直接比較
                if (!matched && aliases && ruby.Name != null && name.Length > 0) {
                    Debug.WriteLine("[DEBUG] Checking Name: {0} <=> {1}", ruby.Name, name);

                    Match nmMatch = Regex.Match(ruby.Name, name, RegexOptions.IgnoreCase);
                    if (nmMatch.Success) {
                        matched = true;
                    }
                } else {
                    Debug.WriteLine("[NOTICE] Skip Direct Name Matching");
                }

                if (matched) {
                    rubies.Add(ruby);
                }
            }
            Debug.WriteLine(String.Format("[DEBUG] Found {0} rubies", rubies.Count));

            return rubies;
        }

        /// <summary>
        /// JSONファイルから設定を読み込みます
        /// </summary>
        /// <param name="path">JSONファイルパス</param>
        /// <returns>成功時true, 失敗時false</returns>
        public bool Load(string path) {
            try {
                var serializer = new DataContractJsonSerializer(typeof(List<RubyProgram>));
                using (FileStream stream = new FileStream(path, FileMode.Open)) {
                    this.Rubies = (List<RubyProgram>) serializer.ReadObject(stream);
                }

                // ソートします
                this.UniqueSort();
                return true;
            } catch (Exception e) {
                Debug.WriteLine(e.ToString());

                return false;
            }
        }

        /// <summary>
        /// JSON形式で設定を書き込みます
        /// </summary>
        /// <param name="path">JSONファイルパス</param>
        /// <returns>成功時true, 失敗時false</returns>
        public bool Save(string path) {
            try {
                this.UniqueSort();
                var serializer = new DataContractJsonSerializer(typeof(List<RubyProgram>));
                using (FileStream stream = new FileStream(path, FileMode.Create)) {
                    serializer.WriteObject(stream, this.Rubies);
                }
                return true;
            } catch (Exception e) {
                Debug.WriteLine(e.ToString());

                return false;
            }
        }

        /// <summary>
        /// リストをソートします
        /// </summary>
        private void UniqueSort() {
            List<RubyProgram> uniqList = new List<RubyProgram>();
            foreach (RubyProgram left in this.Rubies) {
                bool uniq = true;
                foreach (RubyProgram right in uniqList) {
                    if (left.Equals(right)) {
                        uniq = false;
                        break;
                    }
                }
                if (uniq) {
                    uniqList.Add(left);
                }
            }
            this.Rubies = uniqList;
            this.Rubies.Sort(delegate(RubyProgram left, RubyProgram right) {
                if (left == right) {
                    return 0;
                }
                int nameCompare = left.Name.CompareTo(right.Name);
                if (nameCompare != 0) {
                    return nameCompare;
                }

                // 名前が同じ場合はバージョンで比較
                int versionCompare = left.Version.CompareTo(right.Version);
                if (versionCompare != 0) {
                    return versionCompare;
                }

                // アーキテクチャ比較
                int archCompare = left.Architecture.CompareTo(right.Architecture);
                if (archCompare != 0) {
                    return archCompare;
                }

                // プラットフォーム比較
                int platformCompare = left.Platform.CompareTo(right.Platform);
                if (platformCompare != 0) {
                    return platformCompare;
                }

                // ここまで来たらもう同じだよね(´・ω・｀)
                return 0;
            });
            this.Rubies.Reverse();
        }
    }
}
